require "td_model/sql"
require "td_model/errors/failed_to_type_validation_of_column_error"

module TdModel
  class Base
    include ::TdModel::Sql

    attr_accessor :table_name, :attr_values, :where_sql, :sql, :select_sql, :select_keys, :order_sql, :limit_sql

    def initialize(args=nil)
      self.class.set_column(:time) unless self.class.attr_types.keys.include?(:time)
      @attr_values = self.class.attr_default_values.clone
      @select_keys = []
      @attr_values.keys.each do |col|
        proc = Proc.new { @attr_values[col] }
        set_proc = Proc.new { |val| @attr_values.store(col, val) }

        self.class_eval { define_method(col, proc) }
        self.class_eval { define_method("#{col}=", set_proc) }
      end

      unless args.nil?
        args.each do |key, val|
          @attr_values.store(key, val)
        end
      end
    end

    def self.attr_types
      {}
    end

    def self.attr_default_values
      {}
    end

    def self.set_column(name, type=nil)
      attr_types = self.attr_types
      attr_types.store(name, type)
      self.define_singleton_method :attr_types do
        attr_types
      end

      attr_default_values = self.attr_default_values
      attr_default_values.store(name, self.init_type_value(type))
      self.define_singleton_method :attr_default_values do
        attr_default_values
      end
    end

    def self.table_name=(name)
      self.define_singleton_method :table_name do
        name
      end
    end

    def self.table_name
      self.to_s.underscore
    end

    def table_name
      self.class.table_name || self.class.to_s.underscore
    end

    def create
      self.valid!(@attr_values)
      TD.event.post(self.table_name, @attr_values)
    end

    def self.create(args)
      res = self.new(args)
      res.valid!(args)
      TD.event.post(self.table_name, args)
      res
    end

    def valid!(args)
      self.class.attr_types.each do |key, val|
        next if val.nil?
        unless args[key].kind_of?(val)
          raise ::TdModel::Errors::FailedToTypeValidationOfColumnError
        end
      end
    end

    def self.init_type_value(obj)
      case obj.to_s
        when "Integer"
          0
        else
          nil
      end
    end

    def self.where(obj)
      self.new.where(obj)
    end

    def where(obj)
      if obj.instance_of?(Hash)
        obj.each do |key, val|
          if @where_sql.present?
            where = "AND"
          else
            @where_sql = ''
            where = "WHERE"
          end

          if obj.instance_of?(Array)
            @where_sql += " #{where} #{self.table_name}.#{key} IN [#{val.join(',')}]"
          else
            @where_sql += " #{where} #{self.table_name}.#{key} = #{val}"
          end
        end
      elsif obj.instance_of?(String)
        if @where_sql.present?
          where = "AND"
        else
          @where_sql = ''
          where = "WHERE"
        end
        @where_sql += " #{where} #{obj}"
      else
        raise "Error"
      end
      self
    end


    def self.order(col, sort=nil)
      self.new.order(col, sort)
    end

    def order(col, sort=nil)
      if col.instance_of?(Array)
        col.each do |column|
          self.order(column, sort)
        end
      end

      if col.instance_of?(String) || col.instance_of?(Symbol)
        if @order_sql.blank?
          @order_sql = "ORDER BY #{self.table_name}.#{col} #{sort}"
        else
          @order_sql += " ,#{self.table_name}.#{col}"
        end
      else
        raise "ERROR"
      end

      self
    end

    def self.reverse_order(col)
      self.new.reverse_order(col)
    end

    def reverse_order(col)
      self.order(col, 'DESC')
    end

    def self.limit(num=nil)
      self.new.limit(num)
    end

    def limit(num=nil)
      if num.present?
        @limit_sql = "LIMIT #{num}"
      end
      self
    end

    def self.query_execute
      self.new.query_execute
    end

    def query_execute
      if @select_sql.blank?
        @select_keys = []
        self.class.attr_types.keys.each do |key|
          if @select_sql.blank?
            @select_sql = ''
            @select_sql += "SELECT #{self.table_name}.#{key}"
          else
            @select_sql += " ,#{self.table_name}.#{key}"
          end
          @select_keys << key
        end
      end
      @sql = "#{@select_sql} FROM #{self.table_name} #{@where_sql} #{@order_sql} #{@limit_sql}"
      res = ::TdModel::Sql.query(@sql)

      result = []
      res.each do |td_response|
        result_val = {}
        @select_keys.each_with_index do |key, ix|
          result_val.store(key, td_response[ix])
        end
        result << self.class.new(result_val)
      end
      result
    end

    def self.select(args)
      self.new.select(args)
    end

    def select(args)
      @select_sql = ''
      @select_keys = []
      [args].flatten.each do |col|
        if @select_sql.blank?
          @select_sql += "SELECT #{self.table_name}.#{col}"
        else
          @select_sql += " ,#{self.table_name}.#{col}"
        end
        @select_keys << key
      end
      self
    end
  end
end